/*
Copyright (c) 2002, 2003 Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 *****************************************************************************
 * \file uart.c     
 * \brief Functions for serial communication
 *
 *****************************************************************************
 */
 
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/pgmspace.h>

#include "otos_cfg.h"
#include "otos_def.h"
#include "types.h"
#include "uart.h"



unsigned char rx_buf[UART_BUF_SIZE];        /* serial receive buffer */
unsigned char buf_read;                     /* read index, implicitly initialized to 0 */
unsigned char buf_write;                    /* write index, implicitly initialized to 0 */
volatile unsigned char buf_len;             /* number of chars in buffer, implicitly initialized to 0 */


/*!
 *****************************************************************************
 * \brief Send a string from RAM over the serial interface
 * \param string String to send
 *****************************************************************************
 */
void otosPrint(uint8_t *string)
{
    while (*string != 0)    
    {   
        otosPutchar(*(string++));
    }
}


/*!
 *****************************************************************************
 * \brief Send a string from program space over the serial interface
 * \param string String to send
 *****************************************************************************
 */
void otosPrint_P(PGM_P string)
{
    while (PRG_RDB(string) != 0)
    {
        otosPutchar(PRG_RDB(string++));
    }
}

/*!
 *****************************************************************************
 * \brief Send a char over the serial interface
 * \param c Character to send
 *****************************************************************************
 */
void otosPutchar(uint8_t c)
{   
    if (c == '\n')
    {

#ifdef CRLF_OUT
        otosPutchar('\r');
#else
        c = '\r';
#endif

    }
    
#if defined __AVR_ATmega103__
    loop_until_bit_is_set(USR, UDRE);   	/* wait for end of last transmission */
    outb(UDR, c);
#elif defined __AVR_ATmega128__
    loop_until_bit_is_set(UCSR0A, UDRE);	/* wait for end of last transmission */
    outb(UDR0, c);
#elif (defined __AVR_ATmega163__) || (defined __AVR_ATmega323__)
    loop_until_bit_is_set(UCSRA, UDRE); 	/* wait for end of last transmission */
    outb(UDR, c);
#endif
}


/*!
 *****************************************************************************
 * \brief Read a char from the serial interface
 *
 * If the receive buffer is empty, the function waits until a char is received.
 * \return The read character
 *****************************************************************************
 */
uint8_t otosGetchar(void)               /* Read one char from serial port buffer */
{
    uint8_t c;
    uint8_t   istat;    
    
    
    while (otosUartAvail() == 0);     	/* Wait until character has been received  */

    c = rx_buf[buf_read++];             /* Read character  */
    if (buf_read == UART_BUF_SIZE)      /* wrap pointer  */
        buf_read = 0;

    istat = inb(SREG);      /* save cpu status register         */
    cli();                  /* disable interrupts               */
    buf_len--;
    outb(SREG, istat);      /* restore status register (i-bit)  */

    return c;
}



/*!
 *****************************************************************************
 * \brief Check if character are available in receive buffer
 * \return Number of chars in receive buffer
 *****************************************************************************
*/
uint8_t otosUartAvail(void)
{
    uint8_t   istat;    
    uint8_t   avail;


    istat = inb(SREG);      /* save cpu status register         */
    cli();
    avail = buf_len;
    outb(SREG, istat);      /* restore status register (i-bit)  */
    return avail;
}


/*!
 *****************************************************************************
 * \brief Serial Receive Interrupt
 *****************************************************************************
 */
SIGNAL(SIG_UART_RECV)
{
    uint8_t c;
    
#if defined __AVR_ATmega128__
    c = inb(UDR0);
#else    
    c = inb(UDR);
#endif

    if (buf_len == UART_BUF_SIZE)       /* buffer is full, ignore char */
        return;

    rx_buf[buf_write++] = c;            /* store char in ring buffer */
    if (buf_write == UART_BUF_SIZE)     /* wrap pointer */
        buf_write = 0;
    buf_len++;                          /* increase length */
}



/*!
 *****************************************************************************
 * \brief Clear the serial receive buffer
 *
 * The receive interrupt is disabled during this function
 *
 *****************************************************************************
 */
void otosUartClear(void)
{
#if defined __AVR_ATmega103__
    cbi(UCR, RXCIE);        /* disable int */
#elif defined __AVR_ATmega128__
    cbi(UCSR0B, RXCIE);     /* disable int */
#elif (defined __AVR_ATmega163__) || (defined __AVR_ATmega323__)
    cbi(UCSRB, RXCIE);      /* disable int */
#endif
    
    buf_read = 0;
    buf_write = 0;
    buf_len = 0;

#if defined __AVR_ATmega103__
    sbi(UCR, RXCIE);        /* enable int */
#elif defined __AVR_ATmega128__
    cbi(UCSR0B, RXCIE);     /* disable int */
#elif (defined __AVR_ATmega163__) || (defined __AVR_ATmega323__)
    cbi(UCSRB, RXCIE);      /* disable int */
#endif
}

/*!
 *****************************************************************************
 * \brief Returns the latest char in serial receive buffer
 *             if one is available without removing it from
 *             the receive buffer.
 *
 * \return <ul><li> 0: no char available
 *             <li> else: the latest received character
 *        </ul>
 *
 *****************************************************************************
 */
uint8_t otosPeekchar(void)             	/* Read one char from serial port buffer  */
                                        /* without removing it  */
{   
    if (otosUartAvail() == 0)         	/* don't wait for character */
        return 0;

    return rx_buf[buf_read];            /* Read character  */
}

/*!
 *****************************************************************************
 * \brief Read one line. 
 *
 * A line ends on a CR or a CR+LF (depending on definition of macro 'CRLF_IN')
 * or when the maximum number of chars was received.
 *
 * \param  buffer   Buffer for string
 * \param  len      Size of buffer
 *
 * \return          Number of characters stored in buffer
 *                 (without cr, crlf and ending '\0')
 *****************************************************************************
 */
uint8_t otosReadline(uint8_t* buffer, uint8_t len)
{
    uint8_t i = 0;
    uint8_t c;

    do
    {
        c = otosGetchar();
        if (c != '\r')
            buffer[i++] = c;            /* store character */

#ifdef CRLF_IN                          /* lines end with CR+LF */
        else                            /* cr detected */
            otosGetchar();             	/* eat linefeed */
#endif

    } while ((c != '\r') && (i < (len - 1)));

    buffer[i] = 0;

    return i;
}

