/*
Copyright (c) 2002, 2003 Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 *****************************************************************************
 * \file uart.h
 * \brief Header file for uart.c
 *
 *****************************************************************************
 */

#ifndef UART_H
#define UART_H

#include <avr/pgmspace.h>

#include "otos_cfg.h"
#include "types.h"


/*
 * serial communication settings
 */


/* size of the serial receive ring buffer */
#define UART_BUF_SIZE   32

/* #define CRLF_IN, if received lines end with CR+LF */
/* #undef CRLF_IN, if received lines end with CR only */
#undef CRLF_IN

/* #define CRLF_OUT, if sent lines should end with CR+LF */
/* #undef CRLF_OUT, if sent lines should end with CR only */
#undef CRLF_OUT


/*
 * if UART_BAUD_RATE is defined, calculate settings and try to detect
 * invalid or sub-optimal baudrates.
 * otosUartInit() is called without arguments here.
 */
#ifdef UART_BAUD_RATE

    /* Baudrate calculation */
    #define UART_BAUD_1 (F_CPU/(UART_BAUD_RATE*16L)-1)              /* baudrate selector, integer part */
    #define UART_BAUD_1000 (F_CPU /(UART_BAUD_RATE*16L /1000)-1000)
    #if (UART_BAUD_1000 - (UART_BAUD_1 * 1000L)) >= 500             /* > 0.5? */
        #define UART_BAUD_SELECT    (UART_BAUD_1 + 1)               /* round up */
    #else
        #define UART_BAUD_SELECT    (UART_BAUD_1)                   /* else, don't round up */
    #endif

    /* detect invalid baudrates */
    #define UART_BAUD_ERROR ((UART_BAUD_1000 - UART_BAUD_SELECT * 1000L) \
            * 1000L / UART_BAUD_1000)                               /* baudrate error in 1/1000 */

    #if (UART_BAUD_ERROR > 20) || (UART_BAUD_ERROR < -20)           /* more than 2 % difference: error */
        #error The selected baudrate does not work with this crystal frequency (error > 2 %)
    #elif (UART_BAUD_ERROR > 10) || (UART_BAUD_ERROR < -10)         /* more than 1 % difference: warning */
        #warning The selected baudrate is not optimal (error > 1 %)
    #endif


#if defined __AVR_ATmega103__

    #define otosUartInit()                                                    \
        outb(UCR, _BV(RXCIE) | _BV(RXEN) | _BV(TXEN));  /* enable TX and RX */  \
        outb(UBRR, UART_BAUD_SELECT)                    /* set baudrate */

#elif defined __AVR_ATmega128__

    #define otosUartInit()                                                        \
        outb(UCSR0B, _BV(RXCIE) | _BV(RXEN) | _BV(TXEN));   /* enable TX and RX */  \
        outb(UBRR0L, UART_BAUD_SELECT);                     /* set baudrate low */  \
        outb(UBRR0H, 0)                                     /* set baudrate high */

#elif (defined __AVR_ATmega163__) || (defined __AVR_ATmega323__)

    #define otosUartInit()                                                        \
        outb(UCSRB, _BV(RXCIE) | _BV(RXEN) | _BV(TXEN));    /* enable TX and RX */  \
        outb(UBRR, UART_BAUD_SELECT);                       /* set baudrate low */  \
        outb(UBRRH, 0)                                      /* set baudrate high */

#endif

/*
 * if UART_BAUD_RATE is not defined, otosUartInit() must be called with the
 * baudrate as argument
 */
#else

#if defined __AVR_ATmega103__

    /* uart init */
    #define otosUartInit(baud)                                                \
        outb(UCR, _BV(RXCIE) | _BV(RXEN) | _BV(TXEN));  /* enable TX and RX */  \
        outb(UBRR, (F_CPU/(baud*16L)-1))                /* set baudrate */

#elif defined __AVR_ATmega128__

    /* uart init */
    #define otosUartInit(baud)                                                    \
        outb(UCSR0B, _BV(RXCIE) | _BV(RXEN) | _BV(TXEN));   /* enable TX and RX */  \
        outb(UBRR0L, (F_CPU/(baud*16L)-1));                 /* set baudrate */      \
        outb(UBRR0H, 0)                                     /* set baudrate high */

#elif (defined __AVR_ATmega163__) || (defined __AVR_ATmega323__)

    /* uart init */
    #define otosUartInit(baud)                                                    \
        outb(UCSRB, _BV(RXCIE) | _BV(RXEN) | _BV(TXEN));    /* enable TX and RX */  \
        outb(UBRR, (F_CPU/(baud*16L)-1));                   /* set baudrate */      \
        outb(UBRRH, 0)                                      /* set baudrate */

#endif

#endif


/* function prototypes */
void            otosPrint(uint8_t *string);
void            otosPrint_P(PGM_P string);
inline void     otosPutchar(uint8_t);
uint8_t         otosGetchar(void);
uint8_t         otosPeekchar(void);
uint8_t         otosReadline(uint8_t *buffer, uint8_t len);
inline uint8_t  otosUartAvail(void);
void            otosUartClear(void);


#endif
