/*
Copyright (c) 2002, 2003, Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 * \file error.c
 * \brief Error message functions
 *
 */


#include <avr/pgmspace.h>

#include "otos_cfg.h"
#include "otos_def.h"
#include "types.h"
#include "error.h"


/* error handler function pointers */
void (*g_pErrorHandler)(uint8_t*);      /* initialized to NULL by C runtime system */
void (*g_pErrorHandler_P)(PGM_P);       /* initialized to NULL by C runtime system */


/*! 
 * \defgroup error Error Handling
 * Error Handling
 *
 * @{
 */
 

/*!
 *****************************************************************************
 * \brief Handle otOS error
 * 
 * This function will be called by otOS in case of a fatal error.
 * If an error handler was set with otosSetErrorHandler(), this will be
 * called, else the system is halted.
 * otOS applications may call this function, too.
 * \param pMessage Message (from program memory) to be displayed.
 *****************************************************************************
 */
void otosError(uint8_t* pMessage)
{
    if (g_pErrorHandler)
        (*g_pErrorHandler)(pMessage);       /* call error hanlder if available */
    else
        while (1);                          /* else halt system */
}


/*!
 *****************************************************************************
 * \brief Handle otOS error
 * 
 * This function will be called by otOS in case of a fatal error.
 * If an error handler was set with otosSetErrorHandler_P(), this will be
 * called, else the system is halted.
 * otOS applications may call this function, too.
 * \param pMessage Message (from program memory) to be displayed.
 *****************************************************************************
 */
void otosError_P(PGM_P pMessage)
{
    if (g_pErrorHandler_P)
        (*g_pErrorHandler_P)(pMessage);     /* call error hanlder if available */
    else
        while (1);                          /* else halt system */
}


/*!
 *****************************************************************************
 * \brief Set error handler function (data space)
 *
 * This function sets the function, that handles error message strings
 * from data space (RAM).
 * 
 * \param pFunc Pointer to error handler function, that takes a string
 *              from data space
 *****************************************************************************
 */
void otosSetErrorHandler(void (*pFunc)(uint8_t*))
{
    g_pErrorHandler = pFunc;    
}


/*!
 *****************************************************************************
 * \brief Set error handler function (program space)
 *
 * This function sets the function, that handles error message strings
 * from program space (FLASH).
 * 
 * \param pFunc Pointer to error handler function, that takes a string
 *              from program space
 *****************************************************************************
 */
void otosSetErrorHandler_P(void (*pFunc)(PGM_P))
{
    g_pErrorHandler_P = pFunc;
}

/*! @} */
