/*
Copyright (c) 2002, 2003, Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 * \file event.c
 * \brief Event handling functions
 *
 */

/*! 
 * \defgroup event Event Handling
 * Event Handling
 *
 * @{
 */

#include <stdlib.h>
#include <avr/interrupt.h>

#include "otos_cfg.h"
#include "otos_def.h"
#include "types.h"
#include "event.h"
#include "time.h"



/*!
 *****************************************************************************
 * \brief Send event(s) to one or more tasks. If a task waits for them, wake it.
 *
 * \note When called from an interrupt service routine, the scheduler is NOT
 * called after sending the events. It will be called at the next system tick.
 * If you want the scheduler to be called as fast as possible, call
 * otosInterruptReturn() at the end of the ISR.
 * 
 * \param pTask     Task to send the events to. If NULL, the events are sent
 *                  to all tasks in the system.
 * \param events    Event(s) to send
 *****************************************************************************
 */
void otosSendEvent(OtosTask* pTask, uint16_t events)
{
    OtosTask*   pTmp;
    OtosTask*   pNextTask;
    uint8_t     dispatch;

    
    dispatch = ISR_NOT_ACTIVE();    /* call dispatcher only if not called from an isr */
    cli();

    if (pTask != NULL)              /* send event to one tasks */
    {
        otosSendEventToOneTask(pTask, events);
    }
    else                            /* send event to all tasks */
    {
        otosSendEventToOneTask(g_pRunningTask, events);

        pTmp = g_pReadyQueue;
        while (pTmp != NULL)
        {
            otosSendEventToOneTask(pTmp, events);
            pTmp = pTmp->pNext;
        }

        pTmp = g_pBlockedQueue;
        while (pTmp != NULL)
        {
            pNextTask = pTmp->pNext;    /* remember next task, because after wakeup it is not valid any more */
            otosSendEventToOneTask(pTmp, events);
            pTmp = pNextTask;
        }      
    }

    if (dispatch)
        otosScheduler(SCHED_NORM);
    else
        g_performDispatch = TRUE;
}


/*! 
 * \defgroup event_int Internal Event Handling Functions
 * Used internally by otOS, not intended to be called by otOS applications.
 *
 * @{
 */

/*!
 *****************************************************************************
 * \brief Send event(s) to one task. If task waits for them, wake it.
 * \param pTask     Task to send the events to
 * \param events    Event(s) to send
 * \note  Used internally by otOS, not intended to be called by otOS applications.
 *****************************************************************************
 */
void otosSendEventToOneTask(OtosTask* pTask, uint16_t events)
{
    pTask->events |= events;

    /* if task is blocked we must wake it perhaps */
    if (pTask->state == TASK_BLOCKED)
    {
        /* if task is waiting for events */
        if (pTask->waitEvents)
        {
            /* all needed events received? */
            if ((pTask->waitEvents & pTask->events) == pTask->waitEvents)
            {
                pTask->events -= pTask->waitEvents;     /* remove received events */
                pTask->waitEvents = 0;                  /* task doesn't wait for evetns anymore */
                pTask->sleepTicks = 0;                  /* deactivate sleep counter */               
                otosWakeup(pTask);
            }
        }
    }
}

/*! @} */


/*!
 *****************************************************************************
 * \brief   Receive events
 *
 * If the \a events were already received, the function returns immediately.
 * Else it waits the number of otOS timer ticks specified with \a timeout.
 * If \a timeout is 0, the function returns immediately, which allows polling
 * for events.
 * If \a timeout is INFINITE then the function will wait for ever for the events.
 *
 * \param events    Events to receive
 * \param timeout   Number of otOS timer ticks to wait for the events
 *                  <ul> <li> 0:        Don't wait
 *                       <li> INFINITE: Wait infinitely
 *                       <li> else:     Wait number of \a ticks
 *                 </ul>
 *
 * \return  <ul> <li> EVENT_SUCCESS:      The events were received
 *               <li> EVENT_NOT_RECEIVED: The events were not received
 *               <li> EVENT_TIMEOUT:      Timeout expired while waiting for events
 *         </ul>
 *****************************************************************************
 */
uint8_t otosReceiveEvent(uint16_t events, uint32_t timeout)
{

    cli();
    
    /* events already received? */
    if ((events & g_pRunningTask->events) == events)
    {
        g_pRunningTask->events -= events;       /* remove received events */
        sei();
        return EVENT_SUCCESS;
    }

    if (timeout == 0)                           /* don't wait */
    {
        sei();
        return EVENT_NOT_RECEIVED;
    }
    else                                        /* wait */
    {            
        g_pRunningTask->waitEvents = events;
        g_pRunningTask->sleepTicks = (timeout == INFINITE) ? 0 : timeout;
        otosBlock();
        cli();
        if (g_pRunningTask->waitEvents)        /* otosSendEvent must clear this value to indicate event reception */
        {
            g_pRunningTask->waitEvents = 0;
            sei();
            return EVENT_TIMEOUT;
        }
        else
        {
            sei();
            return EVENT_SUCCESS;
        }
    }    
}


/*! @} */
