/*
Copyright (c) 2002, 2003, Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 * \file hardware.c
 * \brief Hardware dependent stuff
 *
 */

/*! 
 * \defgroup hardware Hardware Configuration
 * otOS Hardware Configuration
 *
 * @{
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include "otos_cfg.h"
#include "otos_def.h"
#include "types.h"
#include "hardware.h"


/*!
 *****************************************************************************
 * \brief Init hardware as needed for otOS
 *
 * This function is called by otOS on startup to initialize the hardware.
 * You must change this function to match your hardware.
 * 
 * \note Used internally by otOS, not intended to be called by otOS applications.
 *****************************************************************************
 */

#if defined __AVR_ATmega103__
/*****************************************************************************
 * ATmega103
 */
void otosInitHardware(void)
{
    MCUCR = _BV(SRE);           /* enable external RAM */

#ifdef OTOS_BANKING
    DDRE = 0x3c;                /* bit 2..5 for bank switching */
#endif
    
    /* init timer 0 */

#ifdef OTOS_DEBUG_RTC

    TCCR0 =
            _BV(CS02) |         /* prescaler 256 */
            _BV(CS01) |         /* prescaler 256 */
            _BV(CTC0);          /* clear on compare match */

    OCR0 = 91;                  /* set output compare register */

#else

    ASSR = _BV(AS0);            /* use 32.768 Hz crystal */
    TCCR0 =
            _BV(CS00) |         /* prescaler 1 */
            _BV(CTC0);          /* clear on compare match */

    OCR0 = 127;                 /* set output compare register */

#endif

    TIMSK = _BV(OCIE0);         /* enable output compare interrupt */
    sei();
}


#elif defined __AVR_ATmega128__
/*****************************************************************************
 * ATmega128
 */
void otosInitHardware(void)
{
    
#ifdef OTOS_BANKING
    DDRE = 0x3c;                /* bit 2..5 for bank switching */
#endif
    
    /* init timer 0 */

#ifdef OTOS_DEBUG_RTC

    TCCR0 =
            _BV(CS2) |          /* prescaler 256 */
            _BV(CS1) |          /* prescaler 256 */
            _BV(WGM1);          /* clear on compare match */

    OCR0 = 91;                  /* set output compare register */

#else

    ASSR = _BV(AS0);            /* use 32.768 Hz crystal */
    TCCR0 =
            _BV(CS0) |          /* prescaler 1 */
            _BV(WGM1);          /* clear on compare match */

    OCR0 = 127;                 /* set output compare register */

#endif

    TIMSK = _BV(OCIE0);         /* enable output compare interrupt */
    sei();
}


#elif (defined __AVR_ATmega163__) || (defined __AVR_ATmega323__)
/*****************************************************************************
 * ATmega163 and ATmega323
 */
void otosInitHardware(void)
{
    /* init timer 2 */

#ifdef OTOS_DEBUG_RTC

    TCCR2 =
            _BV(CS22) |         /* prescaler 1 */
            _BV(CS21) |         /* prescaler 1 */
            _BV(CTC2);          /* clear on compare match */

    OCR2 = 91;                  /* set output compare register */

#else

    ASSR = _BV(AS2);            /* use 32.768 Hz crystal */
    TCCR2 =
            _BV(CS20) |         /* prescaler 1 */
            _BV(CTC2);          /* clear on compare match */

    OCR2 = 127;                 /* set output compare register */

#endif

    TIMSK = _BV(OCIE2);         /* enable output compare interrupt */
    sei();
}


#endif

/*! @} */
