/*
Copyright (c) 2002, 2003, Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 * \file memory.c
 * \brief Memory management functions
 *
 */


#include <avr/io.h>
#include <stddef.h>
#include <avr/interrupt.h>

#include "otos_cfg.h"
#include "otos_def.h"
#include "types.h"
#include "task.h"
#include "memory.h"


#ifndef OTOS_BANKING
    #define NUMBER_OF_BANKS 1
#endif

extern size_t __bss_end;


/*
 * check memory configuration
 */

#if XRAMEND > RAMEND                                /* AVR supports external RAM */
    #if EXTERNAL_RAM_END > EXTERNAL_RAM_START       /* available ext. RAM defined and reasonable */

        #define XRAM_AVAILABLE

        #if EXTERNAL_RAM_START == (RAMEND + 1)      /* ext. RAM is contiguous to int. RAM */
            #define XRAM_CONTIGUOUS
        #endif

        #ifndef OTOS_BANKING
            #define USE_XRAM_FOR_HEAP               /* use ext. RAM for heap if no banking is used */
        #endif
    #endif
#endif


static otos_size_t  s_nextFreeInt = (otos_size_t) &__bss_end;   /* first free address in int. RAM */


#if (defined USE_XRAM_FOR_HEAP) && (!defined XRAM_CONTIGUOUS)

static otos_size_t  s_nextFreeExt = EXTERNAL_RAM_START;         /* first free address in ext. RAM */

/*! 
 * \defgroup memory Memory Management
 * Memory Management
 *
 * @{
 */

/*!
 *****************************************************************************
 * \brief Allocate memory from system heap.
 *
 * \param size  Number of bytes to allocate
 * \return      Pointer to allocated memory, or NULL if there is not
 *              enough memory free.
 *****************************************************************************
 */
void* otosAllocate(otos_size_t size)
{
    void*               pMem;
    uint8_t             iStat;
    otos_size_t         blockSize;


    iStat = SREG;                                   /* save cpu status register (i-bit) */
    cli();

    /* try internal RAM */
    blockSize = RAMEND - s_nextFreeInt + 1;
    if (size <= blockSize)                          /* block is big enough */
    {
        pMem = (void*) s_nextFreeInt;
        s_nextFreeInt += size;
    }
    else    
    {
        /* try external RAM */
        blockSize = EXTERNAL_RAM_END - s_nextFreeExt + 1;
        if (size <= blockSize)                      /* block is big enough */
        {
            pMem = (void*) s_nextFreeExt;
            s_nextFreeExt += size;
        }
        else                                        /* block is not big enough */
            pMem = NULL;
    }    

    /* enable interrupts again */
    SREG = iStat;                                   /* restore status register (i-bit) */

    return pMem;
}

#else

/*! 
 * \defgroup memory Memory Management
 * Memory Management
 *
 * @{
 */
 
/*!
 *****************************************************************************
 * \brief Allocate memory from system heap.
 *
 * \param size  Number of bytes to allocate
 * \return      Pointer to allocated memory, or NULL if there is not
 *              enough memory free.
 *****************************************************************************
 */
void* otosAllocate(otos_size_t size)
{
    void*               pMem;
    uint8_t             iStat;
    otos_size_t         blockSize;


    iStat = SREG;                               /* save cpu status register (i-bit) */
    cli();

#if (defined USE_XRAM_FOR_HEAP) && (defined XRAM_CONTIGUOUS)
    blockSize = EXTERNAL_RAM_END - s_nextFreeInt + 1;
#else
    blockSize = RAMEND - s_nextFreeInt + 1;
#endif

    if (size <= blockSize)                      /* block is big enough */
    {
        pMem = (void*) s_nextFreeInt;
        s_nextFreeInt += size;
    }
    else                                        /* block is not big enough */
        pMem = NULL;

    /* enable interrupts again */
    SREG = iStat;                               /* restore status register (i-bit) */

    return pMem;
}

#endif


/*!
 *****************************************************************************
 * \brief Returns the number of free bytes in the heap.
 *
 * \return      Number of free bytes in heap
 *****************************************************************************
 */
otos_size_t otosGetFreeHeap(void)
{
    otos_size_t free;


    free = RAMEND - s_nextFreeInt + 1
#if (defined USE_XRAM_FOR_HEAP)
            + EXTERNAL_RAM_END - s_nextFreeExt + 1
#endif
            ;

    return free;
}


/* this functions are only needed, if we support banking */

#ifdef OTOS_BANKING

/*!
 *****************************************************************************
 * \brief Set RAM bank for external RAM.
 *
 * \param bank  Number of bank [0..NUMBER_OF_BANKS]
 *****************************************************************************
 */
inline void otosSetRamBank(uint8_t bank)
{
    SET_RAM_BANK(bank);
}


/*!
 *****************************************************************************
 * \brief Get active RAM bank.
 *
 * \return  Number of RAM bank [0..NUMBER_OF_BANKS]
 *****************************************************************************
 */
inline uint8_t otosGetRamBank(void)
{
    return GET_RAM_BANK();
}


/*!
 *****************************************************************************
 * \brief Allocate memory from memory bank.
 *
 * \param size  Number of bytes to allocate
 * \param bank  RAM bank
 * \return      Pointer to allocated memory, or NULL if there is not
 *              enough memory free.
 *****************************************************************************
 */
void* otosAllocateBanked(otos_size_t size, uint8_t bank)
{
    static otos_size_t  s_nextFreeAddr[NUMBER_OF_BANKS];
    otos_size_t         mem;
    uint8_t             iStat;


    iStat = SREG;                   /* save cpu status register (i-bit) */
    cli();

    if ((s_nextFreeAddr[bank] + size) <= (EXTERNAL_RAM_END - EXTERNAL_RAM_START + 1))
    {
        mem = s_nextFreeAddr[bank] + EXTERNAL_RAM_START;
        s_nextFreeAddr[bank] += size;
    }
    else
        mem = 0;

    /* enable interrupts again */
    SREG = iStat;                   /* restore status register (i-bit) */

    return (void*) mem;
}

#endif


/*! @} */
