/*
Copyright (c) 2002, 2003, Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 * \file message.c
 * \brief Message queue handling
 *
 */

/*! 
 * \defgroup message Message Queues
 * Message queue handling
 *
 * @{
 */

#include <string.h>
#include <avr/interrupt.h>

#include "otos_cfg.h"
#include "otos_def.h"
#include "types.h"
#include "memory.h"
#include "task.h"
#include "message.h"
#include "time.h"



/*!
 *****************************************************************************
 * \brief Create a message queue.
 *
 * \param   size  Size of the message queue
 *
 * \return  Pointer to created message queue, NULL if otOS is out of memory.
 *****************************************************************************
 */
OtosMessageQueue* otosCreateMessageQueue(uint8_t size)
{
    OtosMessageQueue* pQ;


    /* allocate memory */
    pQ = otosAllocate(sizeof (OtosMessageQueue) + size * sizeof (OtosMessage));
    if (pQ == NULL)
        return NULL;

    /* init queue */
    pQ->size = size;
    pQ->in = 0;
    pQ->out = 0;
    pQ->count = 0;

    return pQ;
}



/*!
 *****************************************************************************
 * \brief Send message to message queue. If a task waits for it, wake it.
 *
 * \note When called from an interrupt service routine, the scheduler is NOT
 * called after sending the message. It will be called at the next system tick.
 * If you want the scheduler to be called as fast as possible, call
 * otosInterruptReturn() at the end of the ISR.
 *
 * \param pQueue Pointer to message queue
 * \param pMsg   Pointer to message
 *
 * \return  <ul> <li> MESSAGE_SUCCESS:     Success
 *               <li> MESSAGE_QUEUE_FULL:  Message queue is full. Message could not be sent.
 *         </ul>
 *****************************************************************************
 */
uint8_t otosSendMessage(OtosMessageQueue* pQueue, OtosMessage* pMsg)
{
    OtosTask*   pTmp;
    OtosTask*   pNextTask;
    uint8_t     dispatch;


    dispatch = ISR_NOT_ACTIVE();   /* call dispatcher only if not called from an isr */
    cli();

    /* message queue full? */
    if (pQueue->count == pQueue->size)
    {
        sei();
        return MESSAGE_QUEUE_FULL;
    }

    /* copy message in queue */
    memcpy(&pQueue->msg[pQueue->in], pMsg, sizeof (OtosMessage));
    MQUEUE_INC_IN(pQueue);

    /* if there is a task waiting on this queue wake it */
    pTmp = g_pBlockedQueue;
    while (pTmp != NULL)
    {
        pNextTask = pTmp->pNext;            /* remember next task, because after wakeup it is not valid any more */

        if (pTmp->pWaitMsgQueue == pQueue)  /* if task waits on this queue */
        {
            pTmp->pWaitMsgQueue = NULL;     /* task doesn't wait for a message anymore */
            pTmp->sleepTicks = 0;           /* deactivate sleep counter */
            otosWakeup(pTmp);
        }

        pTmp = pNextTask;
    }

    if (dispatch)
        otosScheduler(SCHED_NORM);
    else
        g_performDispatch = TRUE;

    return MESSAGE_SUCCESS;
}


/*!
 *****************************************************************************
 * \brief Receive message from message queue.
 *
 * \param pQueue    Message queue pointer
 * \param pMsg      Pointer to message
 * \param timeout   Number of otOS timer ticks to wait for a message
 *                  <ul> <li> 0:        Don't wait, return immediately
 *                       <li> INFINITE: Wait infinitely
 *                       <li> else:     Wait number of ticks
 *                 </ul>
 *
 * \return <ul>
 *   <li> MESSAGE_SUCCESS:       A message was received
 *   <li> MESSAGE_QUEUE_EMPTY:   No message was received, because the queue is empty
 *   <li> MESSAGE_TIMEOUT:       Timeout expired while waiting for a message
 * </ul>
 *
 * \note May not be called from an ISR
 *****************************************************************************
 */
uint8_t otosReceiveMessage(OtosMessageQueue* pQueue, OtosMessage* pMsg, uint32_t timeout)
{
    cli();

    /* if there is a message in the queue, read it */
    if (pQueue->count > 0)
    {
        memcpy(pMsg, &pQueue->msg[pQueue->out], sizeof (OtosMessage));
        MQUEUE_INC_OUT(pQueue);
        sei();
        return MESSAGE_SUCCESS;
    }

    if (timeout == 0)                           /* don't wait */
    {
        sei();
        return MESSAGE_QUEUE_EMPTY;
    }
    else                                        /* wait */
    {
        g_pRunningTask->pWaitMsgQueue = pQueue; /* mark that this task waits for a message on this queue */
        g_pRunningTask->sleepTicks = (timeout == INFINITE) ? 0 : timeout;   /* ticks to wait */
        otosBlock();
        cli();
        if (g_pRunningTask->pWaitMsgQueue)      /* otosSendMessage must clear this value to indicate message reception */
        {
            g_pRunningTask->pWaitMsgQueue = NULL;
            sei();
            return MESSAGE_TIMEOUT;
        }
        else
        {
            memcpy(pMsg, &pQueue->msg[pQueue->out], sizeof (OtosMessage));
            MQUEUE_INC_OUT(pQueue);
            sei();
            return MESSAGE_SUCCESS;
        }
    }
}


/*! @} */
