/*
Copyright (c) 2002, 2003, Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 * \file mutex.c
 * \brief Functions for mutual exclusion
 *
 */

/*! 
 * \defgroup mutex Mutex Handling
 * Mutex Handling
 *
 * @{
 */

#include <avr/interrupt.h>
#include <stddef.h>

#include "otos_cfg.h"
#include "otos_def.h"
#include "types.h"
#include "task.h"
#include "mutex.h"
#include "time.h"



/*!
 *****************************************************************************
 * \brief Test and set a lock (integer variable).
 * 
 * A lock is either locked (1) or unlocked (0).
 * 
 * \param pLock  Pointer to lock
 * \return       The old state of the lock.
 *     <ul> <li> 0: The lock could be locked.
 *          <li> 1: The lock was already locked.
 *    </ul>
 *****************************************************************************
 */
uint8_t otosTestAndSet(uint8_t* pLock)
{
    uint8_t stat;
    uint8_t iStat;    


    iStat = SREG;           /* save cpu status register         */
    cli();
    stat = *pLock;
    *pLock = 1;
    SREG = iStat;           /* restore status register (i-bit)  */
    
    return stat;
}


/*!
 *****************************************************************************
 * \brief Lock a mutex.
 *
 * If the mutex is already locked, the task gets blocked. It is woken up again
 * when the mutex is unlocked by another task.
 * 
 * \param pMutex    Pointer to mutex
 * \param timeout   Number of otOS timer ticks to wait for the mutex
 *                  <ul> <li> 0:        Don't wait, return immediately with
 *                                      status MUTEX_SUCCESS or MUTEX_ALREADY_LOCKED
 *                       <li> INFINITE: Wait infinitely
 *                       <li> else:     Wait number of ticks
 *                 </ul>
 *
 * \return <ul>
 *  <li> MUTEX_SUCCESS:         Mutex was locked successfully
 *  <li> MUTEX_ALREADY_LOCKED:  Mutex was already locked
 *  <li> MUTEX_TIMEOUT:         Timeout while waiting for mutex
 * </ul>
 *****************************************************************************
 */
uint8_t otosLockMutex(OtosMutex* pMutex, uint32_t timeout)
{
    uint8_t lock;

        
    /* test and set mutex */
    lock = otosTestAndSet(&pMutex->lock);       /* lock = 0: we got the mutex. lock = 1: mutex is locked */

    if (lock == 0)                              /* ok, we got the mutex */
    {
        pMutex->pOwner = g_pRunningTask;        /* mark that this is our mutex */
        return MUTEX_SUCCESS;          
    }
    
    if (timeout == 0)                           /* don't wait */
    {
        return MUTEX_ALREADY_LOCKED;            /* could not lock mutex, don't block task */
    }
    else                                        /* wait for mutex */
    {
        cli();
        g_pRunningTask->pWaitMutex = pMutex;
        g_pRunningTask->sleepTicks = (timeout == INFINITE) ? 0 : timeout;
        otosBlock();
        cli();
        if (g_pRunningTask->pWaitMutex)         /* otosUnlockMutex must clear this value to indicate mutex was unlocked */
        {
            g_pRunningTask->pWaitMutex = NULL;
            sei();
            return MUTEX_TIMEOUT;
        }
        else
        {
            pMutex->lock = 1;                   /* lock the mutex */
            pMutex->pOwner = g_pRunningTask;    /* mark that this is our mutex */
            sei();
            return MUTEX_SUCCESS;
        }
    }           
}


/*!
 *****************************************************************************
 * \brief Unlock a mutex.
 *
 * If a task waits for this mutex, it is woken up.
 * 
 * \param pMutex    Pointer to mutex
 *
 * \return <ul>
 *  <li> MUTEX_SUCCESS:         Mutex was unlocked.
 *  <li> MUTEX_NOT_OWNER:       Running task doesn't own this mutex, so it
 *                              cannot unlock it.
 * </ul>
 *****************************************************************************
 */
uint8_t otosUnlockMutex(OtosMutex* pMutex)
{
    OtosTask*   pTmp;
    OtosTask*   pNextTask;
    uint8_t     dispatch;
    uint8_t     unlock = 1;

    if (pMutex->pOwner != g_pRunningTask)   /* task must own the mutex to unlock it */
        return MUTEX_NOT_OWNER;

    dispatch = ISR_NOT_ACTIVE();            /* call dispatcher only if not called from an isr */
    cli();  

    /* wake first waiting task */
    pTmp = g_pBlockedQueue;
    while (pTmp != NULL)
    {
        pNextTask = pTmp->pNext;            /* remember next task, because after wakeup it is not valid any more */

        if (pTmp->pWaitMutex == pMutex)     /* if task waits on this mutex */
        {
            pTmp->pWaitMutex = NULL;        /* task doesn't wait for a mutex anymore */
            pTmp->sleepTicks = 0;           /* deactivate sleep counter */
            pMutex->pOwner = pTmp;          /* give mutex to this task */
            unlock = 0;                     /* don't unlock mutex, because it is now locked by the next task */
            otosWakeup(pTmp);
            break;
        }        
        
        pTmp = pNextTask;
    }

    if (unlock)
        pMutex->lock = 0;                   /* unlock mutex */
    
    if (dispatch)
        otosScheduler(SCHED_NORM);
    else
        g_performDispatch = TRUE;

    return MUTEX_SUCCESS;
}

/*! @} */
