/*
Copyright (c) 2002, 2003, Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 * \file task.c
 * \brief Task management and control functions
 *
 */


#include <avr/interrupt.h>
#include <avr/io.h>
#include <string.h>
#include <avr/pgmspace.h>

#include "otos_cfg.h"
#include "otos_def.h"
#include "types.h"
#include "memory.h"
#include "hardware.h"
#include "error.h"
#include "time.h"
#include "task.h"



/*
 * global variables
 */

uint8_t     msgErrOutOfMem[] PROGMEM = "otOS ERR: out of mem";

/* Note: All global variables are automatically initialized to 0 by the compiler */

/*! \brief All tasks that are ready to run */
OtosTask*   volatile g_pReadyQueue;

/*! \brief All blocked tasks */
OtosTask*   volatile g_pBlockedQueue;

/*! \brief The running task */
OtosTask*   volatile g_pRunningTask;

/*! \brief Total number of tasks */
// uint8_t     g_NumberOfTasks;

/*! \brief TRUE if dispatch must be performed */
uint8_t     volatile g_performDispatch;

/*! \brief Number of timer ticks in a time slice. If 0, disable preemtive multitasking */
uint8_t     volatile g_timeSliceTicks;

/*! \brief Time slice counter */
uint8_t     volatile g_timeSliceCounter;

/* variables used for profiling */
#ifdef OTOS_PROFILING

uint32_t    g_idleEnterTicks;
uint32_t    g_idleTotalTicks;
uint8_t     g_idleActive;

#endif


/*! 
 * \defgroup task Task Management
 * Task Management
 *
 * @{
 */

/*!
 *****************************************************************************
 * \brief Create a task
 *
 * If the created task has a priority higher than the running task, the
 * new task is started immediately, else it is inserted in the ready queue.
 *
 * \param pFunc     Pointer to task function [void pFunc(void)]
 * \param pArg      Arguments that are passed to the task function when it is started
 * \param prio      Task priority [1..250]
 * \param stacksize Size of task's stack in bytes
 * \param name      Task name (one character)
 *
 * \return    Pointer to created task or NULL on error (out of memory)
 *****************************************************************************
 */
OtosTask* otosCreateTask(void (*pFunc)(void*), void* pArg, uint8_t prio, uint16_t stacksize, uint8_t name)
{
    OtosTask*           pTask;
    TaskStackLayout*    pStack;
    uint8_t*            pMem;
    

    /* allocate memory for stack */
    pMem = otosAllocate(stacksize);
    if (pMem == NULL)       /* error: out of memory */
        return NULL;

    /* allocate memory for task structure */
    pTask = otosAllocate(sizeof (OtosTask));
    if (pTask == NULL)      /* error: out of memory */
        return NULL;

#ifdef OTOS_DEBUG_STACK
    /* fill stack with magic pattern */
    memset(pMem, 0x42, stacksize);

    /* store stacksize */
    pTask->stacksize = stacksize;
#endif

#ifdef OTOS_STACK_CHECK
    /* mark stack begin to be able to detect if it overflows */
    pTask->pStack = (uint16_t*) pMem;
    *(pTask->pStack) = STACK_CHECK;
#endif

    /* set stack pointer to last word in stack minus size of TaskStackLayout */
    pStack = (TaskStackLayout*) (pMem + stacksize - sizeof (TaskStackLayout));
//    (void*) pStack += (stacksize - sizeof (TaskStackLayout));

    /* fill stack frame according to GCCs function calling conventions */
    pStack->pcHi = ((uint16_t) pFunc) >> 8;
    pStack->pcLo = ((uint16_t) pFunc) & 0xff;
/*    pStack->r1 = 0;   */
    pStack->r24 = ((uint16_t) pArg) & 0xff;             /* argument */
    pStack->r25 = ((uint16_t) pArg) >> 8;
#ifdef OTOS_BANKING
    pStack->bank = 0;                                   /* the default bank is 0 for every task */
#endif

    cli();

    /* set task control block */
    pTask->sp = (uint16_t) pStack - 1;
    pTask->events = 0;
    pTask->waitEvents = 0;
    pTask->pWaitMsgQueue = NULL;
    pTask->pWaitMutex = NULL;
    pTask->sleepTicks = 0;
    pTask->priority = prio;
    pTask->name = name;

    /* insert task in ready queue */
    otosInsertTaskInQueue(&g_pReadyQueue, pTask, TASK_READY);

    /* update global task variables */
//    ++g_NumberOfTasks;

    /* call scheduler. The new task will be started if it has
     * higher priority than the running task */
    otosScheduler(SCHED_NORM);

    return pTask;
}


/*!
 *****************************************************************************
 * \brief Get pointer to the running task.
 * \return Pointer to running task
 *****************************************************************************
 */
OtosTask* otosGetRunningTask(void)
{
    return g_pRunningTask;
}


/*!
 *****************************************************************************
 * \brief Set the priority of the running task
 *
 * \param prio  New priority
 * \return      The old priority of the running task
 *****************************************************************************
 */
uint8_t otosSetPriority(uint8_t prio)
{
    uint8_t   oldPrio;


    cli();
    oldPrio = g_pRunningTask->priority;
    g_pRunningTask->priority = prio;
    otosScheduler(SCHED_NORM);

    return oldPrio;
}


#ifdef OTOS_PROFILING

/*!
 *****************************************************************************
 * \brief Get the current idle rate
 *
 * The idle rate is the time in percent that the CPU spends in the idle task.
 * \return  Idle rate (0..100)
 *****************************************************************************
 */
uint8_t otosGetIdleRate(void)
{
    return 100 * g_idleTotalTicks / otosGetTicks();
}

#endif



/*! 
 * \defgroup task_int Internal Task Management Functions
 * Used internally by otOS, not intended to be called by otOS applications.
 *
 * @{
 */

/*!
 *****************************************************************************
 * \brief Create the main task. The calling function will become the
 *        main task.
 *
 * \note Used internally by otOS, not intended to be called by otOS applications.
 *
 * \param   prio        Task priority [1..250]
 * \param   stacksize   Size of task's stack in bytes
 *
 * \return  Pointer to the created task structure
 *****************************************************************************
 */
OtosTask* otosCreateMainTask(uint8_t prio, uint16_t stacksize)
{
    OtosTask*   pTask;
    uint8_t*    pSpNew;
    uint16_t    spOld;
    uint8_t*    pMem;


    /* allocate memory for stack */
    pMem = otosAllocate(stacksize);
    if (pMem == NULL)                   /* fatal error, stop system */
        otosError_P(msgErrOutOfMem);

    /* allocate memory for task structure */
    pTask = otosAllocate(sizeof (OtosTask));
    if (pTask == NULL)                  /* fatal error, stop system */
        otosError_P(msgErrOutOfMem);

#ifdef OTOS_DEBUG_STACK
    /* fill stack with magic pattern */
    memset(pMem, 0x42, stacksize);

    /* store stacksize */
    pTask->stacksize = stacksize;
#endif

#ifdef OTOS_STACK_CHECK
    /* mark stack begin to be able to detect if it overflows */
    pTask->pStack = (uint16_t*) pMem;
    *(pTask->pStack) = STACK_CHECK;
#endif

    /* get current stack pointer */
    spOld = GET_SP();

    /* new stack pointer points to last word in stack minus offset of current stack pointer */
    pSpNew = pMem + stacksize - 1  - RAMEND + spOld;

    cli();

    /*
     * now copy stack to new location
     * source:  spOld + 1
     * dest:    pSpNew + 1
     * length:  RAMEND - spOld
     */
    memcpy(pSpNew + 1, (uint8_t*) spOld + 1, RAMEND - spOld);

    /* move stack pointer to new stack */
    __asm__ volatile(
        "out %1, %A0"   "\n\t"
        "out %2, %B0"   "\n\t"
        :
        : "r" (pSpNew), "I" (_SFR_IO_ADDR(SPL)), "I" (_SFR_IO_ADDR(SPH))
    );


    /* set task control block */
    pTask->state = TASK_RUNNING;
    pTask->events = 0;
    pTask->waitEvents = 0;
    pTask->pWaitMsgQueue = NULL;
    pTask->pWaitMutex = NULL;
    pTask->sleepTicks = 0;
    pTask->priority = prio;
    pTask->name = NAME_MAIN;

    /* set global variables */
    g_pRunningTask = pTask;                          /* pointer to the current task */
//    ++g_NumberOfTasks;

    sei();

    return pTask;
}


/*!
 *****************************************************************************
 * \brief Switch task context from \a pOldTask to \a pNewTask
 *
 * \note Used internally by otOS, not intended to be called by otOS applications.
 *
 * \param pOldTask  The running task
 * \param pNewTask  The task to switch to
 *****************************************************************************
 */
void otosDispatch(OtosTask* pOldTask, OtosTask* pNewTask)
{

    /* save context of the old (running) task */
    TASK_CONTEXT_SAVE(pOldTask);

    /* restore context of the new task */
    TASK_CONTEXT_RESTORE(pNewTask);

    /* enable interrupts again */
    sei();
}


/*!
 *****************************************************************************
 * \brief Distribute CPU time.
 *
 * \note Used internally by otOS, not intended to be called by otOS applications.
 *
 * \param operation   <ul> <li>SCHED_BLOCK:  Block running task
 *                         <li>SCHED_NORM:   Normal task switch </ul>
 *****************************************************************************
 */
void otosScheduler(uint8_t operation)
{
    OtosTask* pTmp;
#ifdef OTOS_STACK_CHECK
    static uint8_t   errorMsg[] = "otOS ERR: task x stack overflow";
#endif


    cli();

    g_performDispatch = FALSE;                  /* clear flag */

#ifdef OTOS_STACK_CHECK
    /* test if stack overflowed */
    if (*(g_pRunningTask->pStack) != STACK_CHECK)
    {
        errorMsg[15] = g_pRunningTask->name;    /* replace the x with the task's name */
        otosError(errorMsg);
    }
#endif

    if (operation == SCHED_NORM)                /* running task is moved to ready queue */
    {
        /* if ready queue is empty or if there is no task in ready queue
           with higher or equal prio than running tasks prio, return */
        if ((g_pReadyQueue == NULL) || (g_pReadyQueue->priority < g_pRunningTask->priority))
        {
            sei();
            return;                             /* no task switch */
        }
    }


    /* remember running task */
    pTmp = g_pRunningTask;

    /* select new running task and delete it from ready queue */
    g_pRunningTask = g_pReadyQueue;
    g_pReadyQueue = g_pReadyQueue->pNext;
    g_pRunningTask->state = TASK_RUNNING;

    if (operation == SCHED_NORM)                /* running task is moved to ready queue */
        /* insert old runnning task into ready queue */
        otosInsertTaskInQueue(&g_pReadyQueue, pTmp, TASK_READY);

    else /* if (operation == SCHED_BLOCK) */    /* running task is moved to blocked queue */
        /* insert old runnning task into blocked queue */
        otosInsertTaskInQueue(&g_pBlockedQueue, pTmp, TASK_BLOCKED);

/* profiling */
#ifdef OTOS_PROFILING
    if (!g_idleActive)                          /* idle task not running at the moment */
    {
        if (g_pRunningTask->priority == 0)      /* switch to idle task (only task with prio 0) */
        {
            /* remember idle task entering time */
            g_idleEnterTicks = otosGetTicks();
            g_idleActive = TRUE;
        }
    }
    else                                        /* idle task is currently running */
    {
/*        if (g_pRunningTask->priority != 0) */ /* switch to other task */
        {
            /* remember idle task leaving time */
            g_idleActive = FALSE;
            g_idleTotalTicks += otosGetTickDiff(g_idleEnterTicks);
        }
    }
#endif  /* OTOS_PROFILING */

    /* switch to new task */
    otosDispatch(pTmp, g_pRunningTask);
}


/*!
 *****************************************************************************
 * \brief Wake up task
 *
 * Move \a pTask from blocked queue to ready queue.
 * No task switch will occur.
 *
 * \note Used internally by otOS, not intended to be called by otOS applications.
 *
 * \param pTask  Task to wake up
 *****************************************************************************
 */
void otosWakeup(OtosTask* pTask)
{
    OtosTask* pTmp;
    OtosTask* volatile* pTmpPrev;


    /* point to first entry in blocked queue */
    pTmp = g_pBlockedQueue;             /* points to first OtosTask in queue */
    pTmpPrev = &g_pBlockedQueue;        /* points to first OtosTask* */

    /* find task in queue */
    while ((pTmp != pTask) && (pTmp != NULL))
    {
        pTmpPrev = &pTmp->pNext;
        pTmp = pTmp->pNext;
    }

    if (pTmp == NULL)                   /* fatal error, stop system */
        otosError_P(PSTR("otOS ERR: task not found"));

    /* now remove task from blocked queue*/
    *pTmpPrev = pTask->pNext;

    /* insert task in ready queue */
    otosInsertTaskInQueue(&g_pReadyQueue, pTask, TASK_READY);
}

/*!
 *****************************************************************************
 * \brief Insert task in a queue
 *
 * The otOS task queues are sorted in order of descending priority, i.e.
 * tasks with higher priority are inserted before tasks with lower priority.
 * If more tasks have the same priority, a newly inserted task is inserted
 * after all other tasks of the same priority.
 *
 * \note Used internally by otOS, not intended to be called by otOS applications.
 *
 * \param queue Pointer to queue
 * \param pTask Task to be inserted in queue
 * \param state New task state
 *****************************************************************************
 */
void otosInsertTaskInQueue(OtosTask* volatile* pQueue, OtosTask* pTask, uint8_t state)
{
    OtosTask* pTmp;
    OtosTask* volatile* pTmpPrev;


    /* point to first entry */
    pTmp = *pQueue;                 /* points to first OtosTask in pQueue */
    pTmpPrev = pQueue;              /* points to first OtosTask* */

    /* find first task with lower prio than task */
    while ((pTmp != NULL) && (pTmp->priority >= pTask->priority))
    {
        pTmpPrev = &pTmp->pNext;
        pTmp = pTmp->pNext;
    }

    /* insert task between pTmpPrev and pTmp */
    pTask->pNext = pTmp;
    *pTmpPrev = pTask;

    pTask->state = state;
}

/*!
 *****************************************************************************
 * \brief Enable/disable preemtive multitasking
 *
 * \param  timeSlice:   <ul> <li> 0: No preemtive multitasking
 *                           <li> n: Use time slice of n timer ticks
 *                     </ul>
 * \return Old setting
 *****************************************************************************
 */
uint8_t otosSetPreemtive(uint8_t timeSlice)
{
    uint8_t oldSetting;
    uint8_t iStat;    


    oldSetting = g_timeSliceTicks;

    iStat = SREG;           /* save cpu status register         */
    cli();
    g_timeSliceCounter = timeSlice;
    g_timeSliceTicks = timeSlice;
    SREG = iStat;           /* restore status register (i-bit)  */
    
    return oldSetting;
}

/*!
 *****************************************************************************
 * \brief Get status of preemtive multitasking
 *
 * \return Status of preemptive multitasking
 * \                    <ul> <li> 0: No preemtive multitasking
 *                           <li> n: Use time slice of n timer ticks
 *                     </ul>
 *****************************************************************************
 */
uint8_t otosGetPreemtive(void)
{   
    return g_timeSliceTicks;
}

/*!
 *****************************************************************************
 * \brief The system idle task
 *
 * This task is executed when there is no other task ready to run.
 * This task must always be READY. It must have priority 0.
 *
 * \note Used internally by otOS, not intended to be called by otOS applications.
 *****************************************************************************
 */
OTOS_TASK_FUNCTION(idle, pArg)
{
    while (1)
    {
        otosScheduler(SCHED_NORM);
    }
}


/*!
 *****************************************************************************
 * \brief C entry point.
 *
 * Initializes task handling and finally calls otosMain(),
 * which is a function supplied by the otOS application.
 *
 * \note Used internally by otOS, not intended to be called by otOS applications.
 *****************************************************************************
 */
int main(void)
{
    /* basic hardware init needed for otOS */
    otosInitHardware();

#ifdef OTOS_BANKING
    /* set ram bank 0 */
    otosSetRamBank(0);
#endif

    /* force linker to include time.c (we need the timer ISR) */
    otosGetTicks();

    /* make the calling function the main task */
    otosCreateMainTask(PRIO_MAIN, STACK_MAIN);

    /* create idle task */
    otosCreateTask(idle, NULL, PRIO_IDLE, STACK_IDLE, NAME_IDLE);


    otosMain();                     /* user entry point */
    while (1);                      /* never ever return */
}

/*! @} */


#ifdef OTOS_DEBUG_TASK_QUEUES
#include "dev/uart.h"

void otos_dump_queues(void)
{
    OtosTask* pTask;


    cli();

    otosPrint_P(PSTR("\nRunning Task:  "));
    otosPutchar(g_pRunningTask->name);
    otosPutchar('\n');

    otosPrint_P(PSTR("Ready Queue:   "));
    pTask = g_pReadyQueue;
    while (pTask != NULL)
    {
        otosPutchar(pTask->name);
        otosPutchar(' ');
        pTask = pTask->pNext;
    }
    otosPutchar('\n');

    otosPrint_P(PSTR("Blocked Queue: "));
    pTask = g_pBlockedQueue;
    while (pTask != NULL)
    {
        otosPutchar(pTask->name);
        otosPutchar(' ');
        pTask = pTask->pNext;
    }
    otosPutchar('\n');

    sei();


}
#endif


#ifdef OTOS_DEBUG_STACK
#include "dev/uart.h"

static unsigned char convertBuff[9];

unsigned char hexTab[] PROGMEM = "0123456789abcdef";


/////////////////////////////////////////////////////////////////
// ByteToHex(): convert an unsigned char to a hex string
/////////////////////////////////////////////////////////////////
unsigned char* byteToHex(unsigned char ucValue)
{
    convertBuff[0] = PRG_RDB(&hexTab[ucValue >> 4]);
    convertBuff[1] = PRG_RDB(&hexTab[ucValue & 0x0f]);
    convertBuff[2] = '\0';

    return convertBuff;
}

/////////////////////////////////////////////////////////////////
// ShortToHex(): convert an unsigned short to a hex string
/////////////////////////////////////////////////////////////////
unsigned char* shortToHex(unsigned short usiValue)
{
    convertBuff[0] = PRG_RDB(&hexTab[usiValue >> 12]);
    convertBuff[1] = PRG_RDB(&hexTab[(usiValue >> 8) & 0x0f]);
    convertBuff[2] = PRG_RDB(&hexTab[(usiValue >> 4) & 0x0f]);
    convertBuff[3] = PRG_RDB(&hexTab[usiValue & 0x0f]);
    convertBuff[4] = '\0';

    return convertBuff;
}

/////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////
void otos_dump_task_stack(OtosTask* pTask)
{
    uint8_t*  addr;
    uint16_t  i;


    addr = (uint8_t*) pTask->pStack;

    for (i = 0; i < pTask->stacksize; ++i)
    {
        otosPrint(shortToHex(pTask->stacksize - 1 - i));
        otosPrint(": ");
        otosPrint(byteToHex(*addr));
        otosPrint("\n");
        ++addr;
    }
    otosPrint("\n");
}

void otos_dump_stack(void)
{
    OtosTask* pTask;


    cli();

    otosPrint_P(PSTR("\nRunning Task:\n"));
    otosPutchar(g_pRunningTask->name);
    otosPutchar('\n');
    otos_dump_task_stack(g_pRunningTask);

    otosPrint_P(PSTR("Ready Queue:\n"));
    pTask = g_pReadyQueue;
    while (pTask != NULL)
    {
        otosPutchar(pTask->name);
        otosPutchar('\n');
        otos_dump_task_stack(pTask);
        pTask = pTask->pNext;
    }
    otosPutchar('\n');

    otosPrint_P(PSTR("Blocked Queue:\n"));
    pTask = g_pBlockedQueue;
    while (pTask != NULL)
    {
        otosPutchar(pTask->name);
        otosPutchar('\n');
        otos_dump_task_stack(pTask);
        pTask = pTask->pNext;
    }
    otosPutchar('\n');

    sei();


}


#endif

/*! @} */
