/*
Copyright (c) 2002, 2003, Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 * \file task.h
 * \brief Header file for task.c (Task management and control functions)
 *
 */

#ifndef TASK_H
#define TASK_H


#include <avr/pgmspace.h>

#include "types.h"
#include "otos_cfg.h"
#include "message.h"
#include "hardware.h"


typedef struct otos_task OtosTask;
#include "mutex.h"

/* task structure (tcb, task control block) */
struct otos_task
{
    OtosTask*           pNext;          /* pointer to next task queue entry */
    uint16_t            sp;             /* stack pointer of the task */
    uint8_t             state;          /* task state (running, ready or blocked) */
    uint8_t             priority;       /* task priority (1..250), 250 = highest */
    uint16_t            events;         /* received events */
    uint16_t            waitEvents;     /* events to wait for */
    OtosMessageQueue*   pWaitMsgQueue;  /* message queue on which the task waits for a message */
    OtosMutex*          pWaitMutex;     /* mutex for which the task waits */
    uint32_t            sleepTicks;     /* number of ticks to sleep */
    uint8_t             name;           /* name of the task */
#ifdef OTOS_STACK_CHECK
    uint16_t*           pStack;         /* pointer to first word (lowest address) of task stack */
#endif
#ifdef OTOS_DEBUG_STACK
    uint16_t            stacksize;      /* size of the task stack */
#endif
};


#define OTOS_TASK_FUNCTION(pFunc, pArg) \
void pFunc(void* pArg) __attribute__ ((noreturn)); \
void pFunc(void* pArg)


/*
 * function prototypes
 */

OtosTask*           otosCreateTask(void (*pFunc)(void*), void* pArg, uint8_t prio, uint16_t stacksize, uint8_t name);
inline OtosTask*    otosGetRunningTask(void);
uint8_t             otosSetPriority(uint8_t prio);
void                otosScheduler(uint8_t operation);
extern void         otosMain(void) __attribute__ ((noreturn));
uint8_t             otosSetPreemtive(uint8_t timeSlice);
inline uint8_t      otosGetPreemtive(void);


/* argument constants */
#define SCHED_NORM  0           /* call scheduler normally */
#define SCHED_BLOCK 1           /* block calling task */


/*! 
 * \ingroup task
 * \brief Return from Interrupt.
 *
 * Call this as last instruction in an interrupt service routine
 * if you want to call the otOS scheduler (e. g. if you sent some
 * events to a task and you want to wake it immediately).
 */
#define             otosInterruptReturn()   otosScheduler(SCHED_NORM)

/*! 
 * \ingroup task
 * \brief Block task for ever.
 */
#define             otosBlock()             otosScheduler(SCHED_BLOCK)

#ifdef OTOS_PROFILING
uint8_t             otosGetIdleRate(void);
#endif



#ifdef OTOS_DEBUG_TASK_QUEUES
    void otos_dump_queues(void);
    #define DUMP_QUEUES     otos_dump_queues()
#else
    #define DUMP_QUEUES ((void) 0)
#endif

#ifdef OTOS_DEBUG_STACK
    void otos_dump_task_stack(OtosTask* pTask);
    void otos_dump_stack(void);
#endif    




/*
 * These definitions are only intended for otOS-internal use
 */
#ifdef COMPILE_OTOS

/* task states */
#define TASK_RUNNING    1
#define TASK_READY      2
#define TASK_BLOCKED    3    


#ifdef OTOS_STACK_CHECK
    #define STACK_CHECK 0xaffe
#endif


/* structure of task stack layout. Used for switching tasks */
typedef struct {
#ifdef OTOS_BANKING
    uint8_t bank;
#endif
#ifdef RAMPZ
    uint8_t rampz;
#endif
    uint8_t r31;
    uint8_t r30;
    uint8_t r29;
    uint8_t r28;
/*    uint8_t r27;  */
/*    uint8_t r26;  */
    uint8_t r25;
    uint8_t r24;
/*    uint8_t r23;  */
/*    uint8_t r22;  */
/*    uint8_t r21;  */
/*    uint8_t r20;  */
/*    uint8_t r19;  */
/*    uint8_t r18;  */
    uint8_t r17;
    uint8_t r16;
    uint8_t r15;
    uint8_t r14;
    uint8_t r13;
    uint8_t r12;
    uint8_t r11;
    uint8_t r10;
    uint8_t r9;
    uint8_t r8;
    uint8_t r7;
    uint8_t r6;
    uint8_t r5;
    uint8_t r4;
    uint8_t r3;
    uint8_t r2;
/*    uint8_t r1; */
/*    uint8_t r0; */
    uint8_t pcHi;
    uint8_t pcLo;
} TaskStackLayout;


/* 
 * macros for task switching
 */

/*                                                          
 * We don't save SREG here because we would only need it    
 * if the task switch was triggered by an interrupt. But    
 * interrupt routines save SREG by themselves.              
 * The same is true for R0 (temp reg) and R1 (zero reg).    
 */                                                         
#define TASK_CONTEXT_SAVE_REGS                                  \
    __asm__ volatile (                                          \
        /* registers */                                         \
/*        "push r0"           "\n\t"    */                      \
/*        "push r1"           "\n\t"    */                      \
        "push r2"           "\n\t"                              \
        "push r3"           "\n\t"                              \
        "push r4"           "\n\t"                              \
        "push r5"           "\n\t"                              \
        "push r6"           "\n\t"                              \
        "push r7"           "\n\t"                              \
        "push r8"           "\n\t"                              \
        "push r9"           "\n\t"                              \
        "push r10"          "\n\t"                              \
        "push r11"          "\n\t"                              \
        "push r12"          "\n\t"                              \
        "push r13"          "\n\t"                              \
        "push r14"          "\n\t"                              \
        "push r15"          "\n\t"                              \
        "push r16"          "\n\t"                              \
        "push r17"          "\n\t"                              \
/*         "push r18"          "\n\t"     */                         \
/*         "push r19"          "\n\t"     */                         \
/*         "push r20"          "\n\t"     */                         \
/*         "push r21"          "\n\t"     */                         \
/*         "push r22"          "\n\t"     */                         \
/*         "push r23"          "\n\t"     */                         \
        "push r24"          "\n\t"                              \
        "push r25"          "\n\t"                              \
/*         "push r26"          "\n\t"     */                         \
/*         "push r27"          "\n\t"     */                         \
        "push r28"          "\n\t"                              \
        "push r29"          "\n\t"                              \
        "push r30"          "\n\t"                              \
        "push r31"          "\n\t"                              \
        :                                                       \
        :                                                       \
    )

#define TASK_CONTEXT_SAVE_RAMPZ                                 \
    __asm__ volatile (                                          \
        "in r16, %0"        "\n\t"      /* RAMPZ */             \
        "push r16"          "\n\t"                              \
        :                                                       \
        : "I" (_SFR_IO_ADDR(RAMPZ))                             \
    )

#define TASK_CONTEXT_SAVE_BANK                                  \
    uint8_t sbank;                                              \
    sbank = GET_RAM_BANK();                                     \
    __asm__ volatile (                                          \
        "push %0"           "\n\t"                              \
        :                                                       \
        : "r" (sbank)                                           \
    )

#define TASK_CONTEXT_SAVE_SP(pTask)                             \
    __asm__ volatile (                                          \
        /* save stack pointer */                                \
        "in %A0, %1"        "\n\t"                              \
        "in %B0, %2"        "\n\t"                              \
        : "=r" (pTask->sp)                                      \
        : "I" (_SFR_IO_ADDR(SPL)), "I" (_SFR_IO_ADDR(SPH))      \
    )



#define TASK_CONTEXT_SAVE1(pTask)           \
        TASK_CONTEXT_SAVE_REGS;
        
#ifdef RAMPZ
#define TASK_CONTEXT_SAVE2(pTask)           \
        TASK_CONTEXT_SAVE1(pTask);          \
        TASK_CONTEXT_SAVE_RAMPZ;
#else
#define TASK_CONTEXT_SAVE2(pTask)           \
        TASK_CONTEXT_SAVE1(pTask);
#endif
        
#ifdef OTOS_BANKING
#define TASK_CONTEXT_SAVE3(pTask)           \
        TASK_CONTEXT_SAVE2(pTask);          \
        TASK_CONTEXT_SAVE_BANK;
#else
#define TASK_CONTEXT_SAVE3(pTask)           \
        TASK_CONTEXT_SAVE2(pTask);
#endif

#define TASK_CONTEXT_SAVE(pTask)            \
        TASK_CONTEXT_SAVE3(pTask);          \
        TASK_CONTEXT_SAVE_SP(pTask)




#define TASK_CONTEXT_RESTORE_SP(pTask)                          \
    __asm__ volatile (                                          \
        /* load stack pointer */                                \
        "out %1, %A0"       "\n\t"                              \
        "out %2, %B0"       "\n\t"                              \
        :                                                       \
        : "r" (pTask->sp), "I" (_SFR_IO_ADDR(SPL)),             \
          "I" (_SFR_IO_ADDR(SPH))                               \
    )

#define TASK_CONTEXT_RESTORE_BANK ({                            \
    uint8_t rbank;                                              \
    __asm__ volatile (                                          \
        "pop %0"            "\n\t"      /* get bank */          \
        : "=r" (rbank)                                          \
        :                                                       \
    );                                                          \
    SET_RAM_BANK(rbank);                                        \
})
                                                                
#define TASK_CONTEXT_RESTORE_RAMPZ                              \
    __asm__ volatile (                                          \
        /* load registers */                                    \
        "pop r16"           "\n\t"      /* get RAMPZ */         \
        "out %0, r16"       "\n\t"      /* output */            \
        :                                                       \
        : "I" (_SFR_IO_ADDR(RAMPZ))                             \
    )
        
#define TASK_CONTEXT_RESTORE_REGS                               \
    __asm__ volatile (                                          \
        "pop r31"       "\n\t"                                  \
        "pop r30"       "\n\t"                                  \
        "pop r29"       "\n\t"                                  \
        "pop r28"       "\n\t"                                  \
/*         "pop r27"       "\n\t"       */                           \
/*         "pop r26"       "\n\t"       */                           \
        "pop r25"       "\n\t"                                  \
        "pop r24"       "\n\t"                                  \
/*         "pop r23"       "\n\t"       */                           \
/*         "pop r22"       "\n\t"       */                           \
/*         "pop r21"       "\n\t"       */                           \
/*         "pop r20"       "\n\t"       */                           \
/*         "pop r19"       "\n\t"       */                           \
/*         "pop r18"       "\n\t"       */                           \
        "pop r17"       "\n\t"                                  \
        "pop r16"       "\n\t"                                  \
        "pop r15"       "\n\t"                                  \
        "pop r14"       "\n\t"                                  \
        "pop r13"       "\n\t"                                  \
        "pop r12"       "\n\t"                                  \
        "pop r11"       "\n\t"                                  \
        "pop r10"       "\n\t"                                  \
        "pop r9"        "\n\t"                                  \
        "pop r8"        "\n\t"                                  \
        "pop r7"        "\n\t"                                  \
        "pop r6"        "\n\t"                                  \
        "pop r5"        "\n\t"                                  \
        "pop r4"        "\n\t"                                  \
        "pop r3"        "\n\t"                                  \
        "pop r2"        "\n\t"                                  \
/*        "pop r1"        "\n\t"    */                          \
/*        "pop r0"        "\n\t"    */                          \
        :                                                       \
        :                                                       \
    )


#define TASK_CONTEXT_RESTORE1(pTask)        \
        TASK_CONTEXT_RESTORE_SP(pTask);     \

#ifdef OTOS_BANKING
#define TASK_CONTEXT_RESTORE2(pTask)        \
        TASK_CONTEXT_RESTORE1(pTask);       \
        TASK_CONTEXT_RESTORE_BANK;
#else
#define TASK_CONTEXT_RESTORE2(pTask)        \
        TASK_CONTEXT_RESTORE1(pTask);
#endif

#ifdef RAMPZ
#define TASK_CONTEXT_RESTORE3(pTask)        \
        TASK_CONTEXT_RESTORE2(pTask);       \
        TASK_CONTEXT_RESTORE_RAMPZ;
#else
#define TASK_CONTEXT_RESTORE3(pTask)        \
        TASK_CONTEXT_RESTORE2(pTask);
#endif

#define TASK_CONTEXT_RESTORE(pTask)         \
        TASK_CONTEXT_RESTORE3(pTask);       \
        TASK_CONTEXT_RESTORE_REGS



/*
 * function prototypes
 */
int         main(void);
OtosTask*   otosCreateMainTask(uint8_t prio, uint16_t stacksize);
void        otosInsertTaskInQueue(OtosTask* volatile* pQueue, OtosTask* pTask, uint8_t state);
void        otosDispatch(OtosTask* pOldTask, OtosTask* pNewTask);
void        otosWakeup(OtosTask* pTask);

/* gobal variables (defined in task.c) */

extern OtosTask*  volatile g_pRunningTask;      /* the running task */
extern OtosTask*  volatile g_pReadyQueue;       /* ready queue */
extern OtosTask*  volatile g_pBlockedQueue;     /* blocked queue */
extern uint8_t    volatile g_performDispatch;   /* 1 if dispatch must be performed */

extern uint8_t    volatile g_timeSliceTicks;
extern uint8_t    volatile g_timeSliceCounter;


#endif  /* #ifdef COMPILE_OTOS */

#endif /* TASK_H */
