/*
Copyright (c) 2002, 2003, Oliver Tscherwitschke
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
 * \file time.c
 * \brief Time handling functions
 *
 */


#include <avr/signal.h>
#include <stddef.h>
#include <avr/interrupt.h>

#include "otos_cfg.h"
#include "otos_def.h"
#include "types.h"
#include "task.h"
#include "time.h"



/* the system tick counter */
uint32_t volatile g_ticks; 

/* tick hook function */
void (*g_pTickHook)(void);          /* initialized to NULL by C runtime system */


/*! 
 * \defgroup time Time Management
 * Time Management
 *
 * @{
 */

/*!
 *****************************************************************************
 * \brief Get the system tick counter
 *
 * The system tick counter is incremented TICKS_PER_SEC times per second.
 * \return Number of system ticks elapsed since the start of the system.
 *****************************************************************************
 */
uint32_t otosGetTicks(void)
{
    uint32_t ticks;
    uint8_t  iStat;    


    iStat = SREG;           /* save cpu status register         */
    cli();                  /* disable interrupts               */
    ticks = g_ticks;        /* read tick counter                */
    SREG = iStat;           /* restore status register (i-bit)  */
    
    return ticks;
}


/*!
 *****************************************************************************
 * \brief Block running task for specified number of ticks
 *
 * If \a ticks is zero, immediately call scheduler
 * \param ticks  Number of system ticks to sleep
 *****************************************************************************
 */
void otosSleep(uint32_t ticks)
{
    if (ticks == 0)
        otosScheduler(SCHED_NORM);
    else
    {
        cli();
        g_pRunningTask->sleepTicks = ticks;     /* number of ticks to sleep */
        otosBlock();                            /* block task */        
    }    
}

/*!
 *****************************************************************************
 * \brief Delay running task for specified number of ticks (wait actively, 
 * don't block task!)
 * \param ticks  Number of system ticks to wait
 *****************************************************************************
 */
void otosDelay(uint32_t ticks)
{
    uint32_t startTime;

    
    startTime = otosGetTicks();
    while (otosGetTickDiff(startTime) < ticks);
}


/*!
 *****************************************************************************
 * \brief Set tick hook function
 * 
 * The tick hook function will be called by the system tick handler with a
 * frequency of TICKS_PER_SEC Hz.
 * \param pFunc  Pointer to tick hook function
 *****************************************************************************
 */
void otosSetTickHook(void (*pFunc)(void))
{
    uint8_t  iStat;    


    iStat = SREG;           /* save cpu status register         */
    cli();
    g_pTickHook = pFunc;
    SREG = iStat;           /* restore status register (i-bit)  */
}


/*!
 *****************************************************************************
 * \brief Calculate time difference between an earlier time and now
 * 
 * \param earlier   earlier time (in ticks)
 *****************************************************************************
 */
uint32_t otosGetTickDiff(uint32_t earlier)
{
    uint32_t now = otosGetTicks();
    
    
    if (now >= earlier)
        return now - earlier;
    else
        return 0xffffffff - earlier + now + 1;
}



/*! 
 * \defgroup time_int Internal Time Management Functions
 * Used internally by otOS, not intended to be called by otOS applications.
 *
 * @{
 */

#if (defined __AVR_ATmega103__) || (defined __AVR_ATmega128__)
    #define OTOS_TIMER_IRQ SIG_OUTPUT_COMPARE0
#elif (defined __AVR_ATmega163__) || (defined __AVR_ATmega323__)
    #define OTOS_TIMER_IRQ SIG_OUTPUT_COMPARE2
#endif

/*!
 *****************************************************************************
 * \brief The system timer tick handler
 *
 * This is called TICKS_PER_SEC times per sec
 *****************************************************************************
 */
SIGNAL(OTOS_TIMER_IRQ)
{
    OtosTask* pTask;
    OtosTask* pNextTask;
    

    ++g_ticks;

    /* call tick hook function if specified */
    if (g_pTickHook != NULL)
        (*g_pTickHook)();

    /* check blocked queue if a task must be woken up */
    pTask = g_pBlockedQueue;
    while (pTask != NULL)
    {
        pNextTask = pTask->pNext;           /* remember next task, because after wakeup it is not valid any more */
        if (pTask->sleepTicks != 0)         /* task is sleeping */
        {
            if (--pTask->sleepTicks == 0)   /* dec sleep_ticks and check if it must be woken up */
            {
                otosWakeup(pTask);
                g_performDispatch = TRUE;
            }
        }
        pTask = pNextTask;
    }
    
    /* handle preemtive multitasking */
    if (g_timeSliceTicks)                               /* g_timeSliceTicks != 0: use preemtive m/t */
    {
        if (--g_timeSliceCounter == 0)                  /* time slice expired */
        {
            g_timeSliceCounter = g_timeSliceTicks;
            g_performDispatch = TRUE;                   /* scheduler will do round robin */
        }
    }
    
    /* call scheduler if g_performDispatch is TRUE */
    if (g_performDispatch == TRUE)
        otosScheduler(SCHED_NORM);
}

/*! @} */

/*! @} */
